# =============================================================================
# Ural Facebook-related heuristic functions
# =============================================================================
#
# Collection of functions crafted to work with Facebook's urls.
#
import re
from collections import namedtuple
try:
    from urllib.parse import parse_qs, urljoin, urlsplit, urlunsplit
except ImportError:
    from urlparse import parse_qs, urljoin, urlsplit, urlunsplit

from ural.ensure_protocol import ensure_protocol

BASE_FACEBOOK_URL = 'https://www.facebook.com'

MOBILE_REPLACE_RE = re.compile(r'^([^.]+\.)?facebook\.', re.I)


def convert_facebook_url_to_mobile(url):
    """
    Function parsing the given facebook url and returning the same but for
    the mobile website.
    """
    safe_url = ensure_protocol(url)

    has_protocol = safe_url == url

    scheme, netloc, path, query, fragment = urlsplit(safe_url)

    if 'facebook' not in netloc:
        raise Exception('ural.facebook.convert_facebook_url_to_mobile: %s is not a facebook url' % url)

    netloc = re.sub(MOBILE_REPLACE_RE, 'm.facebook.', netloc)

    result = (
        scheme,
        netloc,
        path,
        query,
        fragment
    )

    result = urlunsplit(result)

    if not has_protocol:
        result = result.split('://', 1)[-1]

    return result


FacebookUser = namedtuple('FacebookUser', ['id', 'handle', 'url'])


def extract_user_from_facebook_url(url):
    if 'facebook.' not in url:
        url = urljoin(BASE_FACEBOOK_URL, url)

    url = ensure_protocol(url)

    parsed = urlsplit(url)

    if parsed.path == '/profile.php':
        query = parse_qs(parsed.query)

        user_id = query['id'][0]
        user_handle = None
        user_url = urljoin(BASE_FACEBOOK_URL, '/profile.php?id=%s' % user_id)
    elif parsed.path.startswith('/people'):
        parts = parsed.path.split('/')

        user_id = parts[3]
        user_handle = None
        user_url = urljoin(BASE_FACEBOOK_URL, '/profile.php?id=%s' % user_id)
    else:
        user_id = None
        user_handle = parsed.path.split('/')[1]
        user_url = urljoin(BASE_FACEBOOK_URL, '/%s' % user_handle)

    return FacebookUser(
        user_id,
        user_handle,
        user_url
    )



from __future__ import unicode_literals
import re

PROTOCOL = r'[a-zA-Z]{0,64}:?//'
WEB_PROTOCOL = r'(?:(?:(?:https?|ftp|wss?):)?//)'

PROTOCOL_RE = re.compile(r'^%s' % PROTOCOL)
WEB_PROTOCOL_RE = re.compile(r'^%s' % WEB_PROTOCOL)

# Taken from:
#  https://gist.github.com/dperini/729294
#  https://gist.github.com/pchc2005/b5f13e136a9c9bb2984e5b92802fc7c9
URL = URL_REGEX = (
    # protocol identifier
    # "(?:(?:(?:https?|ftp):)?//)"
    # user:pass authentication
    r"(?:\S+(?::\S*)?@)?"
    r"(?:"
    # IP address exclusion
    # private & local networks
    r"(?!(?:10|127)(?:\.\d{1,3}){3})"
    r"(?!(?:169\.254|192\.168)(?:\.\d{1,3}){2})"
    r"(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})"
    # IP address dotted notation octets
    # excludes loopback network 0.0.0.0
    # excludes reserved space >= 224.0.0.0
    # excludes network & broadcast addresses
    # (first & last IP address of each class)
    r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
    r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}"
    r"(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))"
    r"|"
    # host & domain names, may end with dot
    # can be replaced by a shortest alternative
    # r"(?![-_])(?:[-\w\u00a1-\uffff]{0,63}[^-_]\.)+"
    # r"(?:(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)"
    # # domain name
    # r"(?:\.(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)*"
    r"(?:"
    r"(?:"
    r"[a-z0-9\u00a1-\uffff]"
    r"[a-z0-9\u00a1-\uffff_-]{0,62}"
    r")?"
    r"[a-z0-9\u00a1-\uffff]\."
    r")+"
    # TLD identifier name, may end with dot
    r"(?:[a-z\u00a1-\uffff]{2,}\.?)"
    r")"
    # port number (optional)
    r"(?::\d{2,5})?"
    # resource path (optional)
    # r"(?:[/?#]\S*)?"
)

RESOURCE_PATH = r"(?:[/?#]\S*)?"
RELAXED_RESOURCE_PATH = r"(?:[/?#][\S ]*)?"

URL_RE = re.compile(
    r'^(?:%s)?%s$' % (PROTOCOL, URL + RESOURCE_PATH), re.I | re.UNICODE)

URL_WITH_PROTOCOL_RE = re.compile(
    r'^%s%s$' % (PROTOCOL, URL + RESOURCE_PATH), re.I | re.UNICODE)

RELAXED_URL = re.compile(
    r'^(?:%s)?%s$' % (PROTOCOL, URL + RELAXED_RESOURCE_PATH), re.I | re.UNICODE)

RELAXED_URL_WITH_PROTOCOL_RE = re.compile(
    r'^%s%s$' % (PROTOCOL, URL + RELAXED_RESOURCE_PATH), re.I | re.UNICODE)

URL_IN_TEXT_RE = re.compile(
    r'(%s)%s' % (PROTOCOL, URL + RESOURCE_PATH), re.I | re.UNICODE)

URL_IN_HTML_RE = re.compile(
    r"<a\s.*?href=(?:\"([.#]+?)\"|\'([.#]+?)\'|([^\s]+?))(?:>|\s.*?>)(?:.*?)<[/ ]?a>",
    re.DOTALL | re.IGNORECASE)



# =============================================================================
# Ural CLI Utils
# =============================================================================
#
# Miscellaneous helpers used by the CLI tools.
#
import csv


def custom_reader(f, target_header):

    reader = csv.reader(f)

    headers = next(reader, None)

    position = headers.index(target_header)

    return headers, position, reader



# =============================================================================
# Ural LRU from URL Function
# =============================================================================
#
# A function returning the url parts in hierarchical order.
#
try:
    from urllib.parse import urlsplit
except ImportError:
    from urlparse import urlsplit

from ural.ensure_protocol import ensure_protocol


def parsed_url_to_lru(parsed_url):
    scheme, netloc, path, query, fragment = parsed_url
    lru = []
    if scheme:
        lru.append('s:' + scheme)

    # Parsing domain & port
    netloc = netloc.split(':')
    if len(netloc) == 2:
        port = netloc[1]
        lru.append('t:' + port)
    for element in reversed(netloc[0].split('.')):
        lru.append('h:' + element)

    # Parsing the path
    for element in path.split('/')[1:]:
        lru.append('p:' + element)
    if query and query[0]:
        lru.append('q:' + query)
    if fragment and fragment[0]:
        lru.append('f:' + fragment)

    return lru


def lru_from_url(url, default_protocol='http'):
    """
    Function returning the parts of the given url in the hierarchical order (lru).

    Args:
        url (str): Target URL as a string.
        default_protocol (str, optional): Protocol to add if there is none.
            Defaults to `'http'`.

    Returns:
        list: The lru, with a prefix identifying the type of each part.
    """

    full_url = ensure_protocol(url, protocol=default_protocol)
    return parsed_url_to_lru(urlsplit(full_url))



# =============================================================================
# Ural Url testing Function
# =============================================================================
#
# A function returning True if its argument is a url.
#
from tld import get_tld
from ural.patterns import (
    URL_RE,
    URL_WITH_PROTOCOL_RE,
    RELAXED_URL,
    RELAXED_URL_WITH_PROTOCOL_RE
)


def is_url(string, require_protocol=True, tld_aware=False,
           allow_spaces_in_path=False):
    """
    Function returning True if its string argument is a url.

    Args:
        string (str): string to test.
        require_protocol (bool, optional): whether the argument has to have a
            protocol to be considered a url. Defaults to True.
        tld_aware (bool, optional): whether to check whether the url's tld
            exists. Defaults to False.
        allow_spaces_in_path (bool, optional): whether to accept spaces in
            the url's path. Defaults to False.

    Returns:
        bool: True if the argument is a url, False if not.

    """
    string = string.strip()

    if require_protocol:
        if allow_spaces_in_path:
            pattern = RELAXED_URL_WITH_PROTOCOL_RE
        else:
            pattern = URL_WITH_PROTOCOL_RE
    else:
        if allow_spaces_in_path:
            pattern = RELAXED_URL
        else:
            pattern = URL_RE

    if not string:
        return False

    if not pattern.match(string):
        return False

    if tld_aware and get_tld(string, fail_silently=True, fix_protocol=True) is None:
        return False

    return True



# =============================================================================
# Ural Normalized LRU from URL Function
# =============================================================================
#
# A function normalizing the url and returning its parts in the hierarchical order.
#
from ural.ensure_protocol import ensure_protocol
from ural.lru_from_url import parsed_url_to_lru
from ural.normalize_url import normalize_url


def normalized_lru_from_url(url, default_protocol='http', **kwargs):
    """
    Function normalizing the given url by stripping it of usually
    non-discriminant parts such as irrelevant query items or sub-domains, and
    returning its parts in the hierarchical order (lru).

    Args:
        url (str): Target URL as a string.
        sort_query (bool, optional): Whether to sort query items or not.
            Defaults to `True`.
        strip_authentication (bool, optional): Whether to drop authentication.
            Defaults to `True`.
        strip_trailing_slash (bool, optional): Whether to drop trailing slash.
            Defaults to `False`.
        strip_index (bool, optional): Whether to drop trailing index at the end
            of the url. Defaults to `True`.

    Returns:
        list: The normalized lru, with a prefix identifying the type of each part.
    """

    full_url = ensure_protocol(url, protocol=default_protocol)
    return parsed_url_to_lru(normalize_url(
        full_url, parsed=True, **kwargs))



#!/usr/bin/env python
# =============================================================================
# Ural CLI Endpoint
# =============================================================================
#
# CLI Interface of the Ural library.
#
import sys
from argparse import ArgumentParser, FileType

from ural.cli.domain import domain_action
from ural.cli.normalize import normalize_action
from ural.cli.join import join_action


SUBPARSERS = {}


def main():
    parser = ArgumentParser(prog='ural')
    subparsers = parser.add_subparsers(
        help='action to execute', title='actions', dest='action')

    # Normalize action subparser
    normalize_subparser = subparsers.add_parser(
        'normalize', description='Normalize the urls of a given CSV column.')
    normalize_subparser.add_argument('column', help='column')
    normalize_subparser.add_argument(
        'file', help='csv file containing the urls to normalize', type=FileType('r'), default=sys.stdin, nargs='?')
    normalize_subparser.add_argument(
        '-o', '--output', help='output file', type=FileType('w'), default=sys.stdout)
    normalize_subparser.add_argument(
        '--no-query-sort', action='store_true', help='do not sort query items', default=False)
    normalize_subparser.add_argument(
        '--keep-authentication', action='store_true', help='do not drop authentication', default=False)
    normalize_subparser.add_argument(
        '--strip-trailing-slash', action='store_true', help='drop trailing slash', default=False)
    normalize_subparser.add_argument(
        '--keep-index', action='store_true', help='do not drop trailing index at the end of the url', default=False)
    SUBPARSERS['normalize'] = normalize_subparser

    # Domain action subparser
    domain_subparser = subparsers.add_parser(
        'domain', description='Extract the urls\' domains of a given CSV column.')
    domain_subparser.add_argument('column', help='column')
    domain_subparser.add_argument(
        'file', help='csv file containing the urls', type=FileType('r'), default=sys.stdin, nargs='?')
    domain_subparser.add_argument(
        '-o', '--output', help='output file', type=FileType('w'), default=sys.stdout)
    SUBPARSERS['normalize'] = domain_subparser

    # Join action subparser
    join_subparser = subparsers.add_parser(
        'join', description='Join 2 csv files according to their urls columns.')
    join_subparser.add_argument(
        'column1', help='name of the url column of the first file')
    join_subparser.add_argument(
        'file1', help='first csv file', type=FileType('r'), default=sys.stdin, nargs='?')
    join_subparser.add_argument(
        'column2', help='name of the url column of the second file')
    join_subparser.add_argument(
        'file2', help='second csv file', type=FileType('r'), default=sys.stdin, nargs='?')
    join_subparser.add_argument(
        '-o', '--output', help='output file', type=FileType('w'), default=sys.stdout)
    join_subparser.add_argument(
        '-s', '--select', nargs='+', help='columns of the first file wanted in the output')
    join_subparser.add_argument(
        '--large-cells', action='store_true', help='activate if the csv contains huge cells (>131072)', default=False)
    SUBPARSERS['join'] = join_subparser

    # Help action subparser
    help_suparser = subparsers.add_parser('help')
    help_suparser.add_argument('subcommand', help='name of the subcommand')
    SUBPARSERS['help'] = help_suparser

    args = parser.parse_args()

    if args.action == 'help':
        target_subparser = SUBPARSERS.get(args.subcommand)

        if target_subparser is None:
            parser.print_help()
        else:
            target_subparser.print_help()

    elif args.action == 'normalize':
        normalize_action(args)

    elif args.action == 'domain':
        domain_action(args)

    elif args.action == 'join':
        join_action(args)

    elif args.action is None:
        parser.print_help()


if __name__ == '__main__':
    main()



# =============================================================================
# Ural LRUTrie class
# =============================================================================
#
# A class implementing a prefix tree(Trie) storing LRUs and their metadata,
# allowing to find the longest common prefix between two urls.
#
from phylactery import TrieDict
from ural.normalized_lru_from_url import normalized_lru_from_url
from functools import partial


class LRUTrie(object):
    def __init__(self, **kwargs):
        self.trie = TrieDict()
        self.normalize = partial(normalized_lru_from_url, **kwargs)

    def set(self, url, metadata):
        lru = self.normalize(url)
        self.trie.set(lru, metadata)

    def match(self, url):
        lru = self.normalize(url)
        return self.trie.longest(lru)

    def values(self):
        return self.trie.values()

    def __iter__(self):
        return self.trie.values()



# =============================================================================
# Ural Normalize CLI Action
# =============================================================================
#
# Logic of the normalize CLI action enabling the user to normalize the urls of
# a CSV file column.
#
import csv

from ural.cli.utils import custom_reader
from ural import normalize_url


def normalize_action(namespace):
    sort_query = not namespace.no_query_sort
    strip_authentication = not namespace.keep_authentication
    strip_trailing_slash = namespace.strip_trailing_slash
    strip_index = not namespace.keep_index

    headers, position, reader = custom_reader(namespace.file, namespace.column)

    headers.append(namespace.column + "_normalized")
    writer = csv.writer(namespace.output)
    writer.writerow(headers)

    for line in reader:
        url = line[position]
        line.append(normalize_url(url, sort_query=sort_query, strip_authentication=strip_authentication,
                                  strip_trailing_slash=strip_trailing_slash, strip_index=strip_index))
        writer.writerow(line)



# =============================================================================
# Ural URL Force Protocol Function
# =============================================================================
#
# A function force-replacing the url protocol by the given one (and adding it if there is none).
#
import re

from ural.patterns import PROTOCOL_RE


def force_protocol(url, protocol='http'):
    """
    Function force-replacing the url protocol by the given one (and adding it if there is none).

    Args:
        url (str): Target URL as a string.
        protocol (str): protocol wanted. Is 'http' by default.

    Returns:
        string: The protocol-equipped url.

    """
    protocol = protocol.rstrip(':/')

    if not PROTOCOL_RE.match(url):
        url = protocol + '://' + url
    elif url[:2] == '//':
        url = protocol + ':' + url
    else:
        url = re.sub(PROTOCOL_RE, protocol + '://', url)

    return url



# =============================================================================
# Ural Join CLI Action
# =============================================================================
#
# Logic of the join CLI action enabling the user to join 2 csv files according
# to the urls each of them contains.
#
import csv
import sys

from ural.cli.utils import custom_reader
from ural.lru_trie import LRUTrie


def join_action(namespace):

    if namespace.large_cells:
        csv.field_size_limit(sys.maxsize)
    file1_headers, file1_position, file1_reader = custom_reader(
        namespace.file1, namespace.column1)
    file2_headers, file2_position, file2_reader = custom_reader(
        namespace.file2, namespace.column2)
    if namespace.select:
        headers = namespace.select + file2_headers
    else:
        headers = file1_headers + file2_headers
    writer = csv.writer(namespace.output)
    writer.writerow(headers)

    trie = LRUTrie()

    for line in file1_reader:
        url = line[file1_position]
        if namespace.select:
            try:
                metadata = [line[file1_headers.index(
                    x)] for x in namespace.select]
            except ValueError as e:
                print("Woops, the header '" + str(e)
                      [1:-16] + "' doesn't exist !")
                sys.exit(1)
        else:
            metadata = line
        trie.set(url, metadata)

    for line in file2_reader:
        metadata = line
        trie_metadata = trie.match(line[file2_position])
        if trie_metadata:
            row = trie_metadata + metadata
        elif namespace.select:
            row = ['' for i in range(len(namespace.select))] + metadata
        else:
            row = ['' for i in range(len(file1_headers))] + metadata
        writer.writerow(row)



# =============================================================================
# Ural Domain Name Getter
# =============================================================================
#
# Function returning an url's domain name.
#
from tld import get_fld


def get_domain_name(url):
    """
    Function returning the given url's domain name as a string. This function
    is of course TLD aware.

    Args:
        url (string): Target url.

    Returns:
        str: the url's domain name.

    """
    return get_fld(url, fix_protocol=True, fail_silently=True)



# =============================================================================
# Ural URL Strip Protocol Function
# =============================================================================
#
# A function removing the protocol from the given url.
#
from ural.patterns import PROTOCOL_RE


def strip_protocol(url):
    """
    Function removing the protocol from the given url.

    Args:
        url (str): Target URL as a string.

    Returns:
        string: The url without protocol.

    """
    return PROTOCOL_RE.sub("", url)



# =============================================================================
# Ural URL Extraction Function
# =============================================================================
#
# A function returning an iterator over the urls present in the string argument.
#
from __future__ import unicode_literals
import re
from ural.patterns import URL_IN_TEXT_RE


def urls_from_text(string):
    """
    Function returning an iterator over the urls present in the string argument.

    Args:
        string (str): source string.

    Yields:
        str: an url.

    """
    for url in re.finditer(URL_IN_TEXT_RE, string):
        yield url.group(0)



# =============================================================================
# Ural URL Ensure Protocol Function
# =============================================================================
#
# A function checking if the url has a protocol, and adding one if there is none.
#
from ural.patterns import PROTOCOL_RE


def ensure_protocol(url, protocol='http'):
    """
    Function checking if the url has a protocol, and adding the given one if there is none.

    Args:
        url (str): Target URL as a string.
        protocol (str): protocol to use if there is none in url. Is 'http' by default.

    Returns:
        string: The protocol-equipped url.

    """
    protocol = protocol.rstrip(':/')

    if not PROTOCOL_RE.match(url):
        url = protocol + '://' + url
    elif url.startswith('//'):
        url = protocol + ':' + url

    return url



# =============================================================================
# Ural Domain CLI Action
# =============================================================================
#
# Logic of the domain CLI action enabling the user to extract the domain name
# of urls contained in a CSV column.
#
import csv

from ural.cli.utils import custom_reader
from ural import get_domain_name


def domain_action(namespace):
    headers, position, reader = custom_reader(namespace.file, namespace.column)

    headers.append(namespace.column + "_domain")
    writer = csv.writer(namespace.output)
    writer.writerow(headers)

    for line in reader:
        url = line[position]
        line.append(get_domain_name(url))
        writer.writerow(line)



from setuptools import setup, find_packages

with open('./README.md', 'r') as f:
    long_description = f.read()

setup(name='ural',
      version='0.11.0',
      description='A helper library full of URL-related heuristics.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='http://github.com/medialab/ural',
      license='MIT',
      author='Guillaume Plique, Jules Farjas, Oubine Perrin, Benjamin Ooghe-Tabanou',
      author_email='kropotkinepiotr@gmail.com',
      keywords='url',
      python_requires='>=2.7',
      packages=find_packages(exclude=['test']),
      package_data={'docs': ['README.md']},
      install_requires=[
        'phylactery>=0.2.2',
        'pycountry>=18.12.8',
        'tld>=0.9.3'
      ],
      entry_points={
        'console_scripts': ['ural=ural.cli.__main__:main']
      },
      zip_safe=True)



# =============================================================================
# Ural URL Extraction From HTML Function
# =============================================================================
#
# A function returning an iterator over the urls present in the HTML string
# argument.
#
from __future__ import unicode_literals
import re
from ural.patterns import URL_IN_HTML_RE


def clean_link(link):
    """Removes leading and trailing whitespace and punctuation"""
    return link.strip("\t\r\n '\"\x0c")


def urls_from_html(string):
    """
    Function returning an iterator over the urls present in the HTML string argument.

    Args:
        string (str): source html string.

    Yields:
        str: an url.

    """
    for a_tag in re.finditer(URL_IN_HTML_RE, string):
        url = a_tag.group(3)
        yield clean_link(url)



# =============================================================================
# Ural Library Endpoint
# =============================================================================
#
from ural.ensure_protocol import ensure_protocol
from ural.force_protocol import force_protocol
from ural.get_domain_name import get_domain_name
from ural.is_url import is_url
from ural.lru_trie import LRUTrie
from ural.lru_from_url import lru_from_url
from ural.normalize_url import normalize_url
from ural.normalized_lru_from_url import normalized_lru_from_url
from ural.strip_protocol import strip_protocol
from ural.urls_from_text import urls_from_text
from ural.urls_from_html import urls_from_html



# =============================================================================
# Ural URL Normalization Function
# =============================================================================
#
# A handy function relying on heuristics to find and drop irrelevant or
# non-discriminant parts of a URL.
#
import re
import pycountry
from os.path import normpath, splitext
try:
    from urllib.parse import parse_qsl, urlsplit, urlunsplit
except ImportError:
    from urlparse import parse_qsl, urlsplit, urlunsplit

from ural.patterns import PROTOCOL_RE

IRRELEVANT_QUERY_RE = re.compile(
    r'^(?:__twitter_impression|echobox|fbclid|fref|utm_.+|amp_.+|amp|s?een|xt(?:loc|ref|cr|np|or|s))$', re.I)

IRRELEVANT_SUBDOMAIN_RE = re.compile(r'\b(?:www\d?|mobile|m)\.', re.I)

IRRELEVANT_QUERY_COMBOS = {
    'ref': ('fb', 'ts', 'tw', 'tw_i'),
    'platform': ('hootsuite')
}


def attempt_to_decode_idna(string):
    try:
        return string.encode('utf8').decode('idna')
    except:
        return string


def stringify_qs(item):
    if item[1] == '':
        return item[0]

    return '%s=%s' % item


def should_strip_query_item(item):
    key = item[0].lower()

    if IRRELEVANT_QUERY_RE.match(key):
        return True

    value = item[1]

    if key in IRRELEVANT_QUERY_COMBOS:
        return value in IRRELEVANT_QUERY_COMBOS[key]

    return False


def normalize_url(url, parsed=False, sort_query=True, strip_authentication=True,
                  strip_trailing_slash=False, strip_index=True, strip_protocol=True,
                  strip_irrelevant_subdomain=True, strip_lang_subdomains=False,
                  strip_fragment='except-routing'):
    """
    Function normalizing the given url by stripping it of usually
    non-discriminant parts such as irrelevant query items or sub-domains etc.

    This is a very useful utility when attempting to match similar urls
    written slightly differently when shared on social media etc.

    Args:
        url (str): Target URL as a string.
        sort_query (bool, optional): Whether to sort query items or not.
            Defaults to `True`.
        strip_authentication (bool, optional): Whether to drop authentication.
            Defaults to `True`.
        strip_trailing_slash (bool, optional): Whether to drop trailing slash.
            Defaults to `False`.
        strip_index (bool, optional): Whether to drop trailing index at the end
            of the url. Defaults to `True`.
        strip_lang_subdomains (bool, optional): Whether to drop language subdomains
            (ex: 'fr-FR.lemonde.fr' to only 'lemonde.fr' because 'fr-FR' isn't a relevant subdomain, it indicates the language and the country).
            Defaults to `False`.
        strip_fragment (bool|str, optional): Whether to drop non-routing fragment from the url?
            If set to `except-routing` will only drop non-routing fragment (i.e. fragments that
            do not contain a "/").
            Defaults to `except-routing`.

    Returns:
        string: The normalized url.

    """

    has_protocol = PROTOCOL_RE.match(url)
    # Ensuring scheme so parsing works correctly
    if not has_protocol:
        url = 'http://' + url

    # Parsing
    scheme, netloc, path, query, fragment = urlsplit(url)

    # Handling punycode
    if 'xn--' in netloc:
        netloc = '.'.join(
            attempt_to_decode_idna(x) for x in netloc.split('.')
        )

    # Dropping :80 & :443
    if netloc.endswith(':80'):
        netloc = netloc[:-3]
    elif netloc.endswith(':443'):
        netloc = netloc[:-4]

    # Normalizing the path
    if path:
        trailing_slash = False
        if path.endswith('/') and len(path) > 1:
            trailing_slash = True
        path = normpath(path)
        if trailing_slash and not strip_trailing_slash:
            path = path + '/'

    # Dropping index:
    if strip_index:
        segments = path.rsplit('/', 1)

        if len(segments) != 0:
            last_segment = segments[-1]
            filename, ext = splitext(last_segment)

            if filename == 'index':
                segments.pop()
                path = '/'.join(segments)

    # Dropping irrelevant query items
    if query:
        qsl = parse_qsl(query, keep_blank_values=True)
        qsl = [
            stringify_qs(item)
            for item in qsl
            if not should_strip_query_item(item)
        ]

        if sort_query:
            qsl = sorted(qsl)

        query = '&'.join(qsl)

    # Dropping fragment if it's not routing
    if fragment and strip_fragment:
        if strip_fragment is True or '/' not in fragment:
            fragment = ''

    # Dropping irrelevant subdomains
    if strip_irrelevant_subdomain:
        netloc = re.sub(
            IRRELEVANT_SUBDOMAIN_RE,
            '',
            netloc
        )

    # Dropping language as subdomains
    if strip_lang_subdomains:
        if netloc.count('.') > 1:
            subdomain, remaining_netloc = netloc.split('.', 1)
            if len(subdomain) == 5 and '-' in subdomain:
                lang, country = subdomain.split('-', 1)
                if len(lang) == 2 and len(country) == 2:
                    if pycountry.countries.get(alpha_2=lang.upper()) and pycountry.countries.get(alpha_2=country.upper()):
                        netloc = remaining_netloc
            elif len(subdomain) == 2:
                if pycountry.countries.get(alpha_2=subdomain.upper()):
                    netloc = remaining_netloc

    # Dropping scheme
    if strip_protocol or not has_protocol:
        scheme = ''

    # Dropping authentication
    if strip_authentication:
        netloc = netloc.split('@', 1)[-1]

    # Dropping trailing slash
    if strip_trailing_slash and path.endswith('/'):
        path = path.rstrip('/')

    # Result
    result = (
        scheme,
        netloc.lower(),
        path,
        query,
        fragment
    )

    if parsed:
        return result

    # TODO: check if works with `parsed=True`
    if strip_protocol or not has_protocol:
        result = urlunsplit(result)[2:]
    else:
        result = urlunsplit(result)

    return result
